package sun.misc;

import java.util.NoSuchElementException;
import java.util.TreeSet;
import java.util.Enumeration;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.IOException;
import java.util.Set;
import java.util.List;
import java.io.BufferedReader;
import java.net.URL;

public final class Service<S>
{
    private static final String prefix = "META-INF/services/";
    
    private Service() {
    }
    
    private static void fail(Class<?> clazz, String str, Throwable cause) throws ServiceConfigurationError {
        ServiceConfigurationError serviceConfigurationError = new ServiceConfigurationError(clazz.getName() + ": " + str);
        serviceConfigurationError.initCause(cause);
        throw serviceConfigurationError;
    }
    
    private static void fail(Class<?> clazz, String str) throws ServiceConfigurationError {
        throw new ServiceConfigurationError(clazz.getName() + ": " + str);
    }
    
    private static void fail(Class<?> clazz, URL obj, int i, String str) throws ServiceConfigurationError {
        fail(clazz, obj + ":" + i + ": " + str);
    }
    
    private static int parseLine(Class<?> clazz, URL url, BufferedReader bufferedReader, int n, List<String> list, Set<String> set) throws IOException, ServiceConfigurationError {
        String s = bufferedReader.readLine();
        if (s == null) {
            return -1;
        }
        int index = s.indexOf(35);
        if (index >= 0) {
            s = s.substring(0, index);
        }
        String trim = s.trim();
        int length = trim.length();
        if (length != 0) {
            if (trim.indexOf(32) >= 0 || trim.indexOf(9) >= 0) {
                fail(clazz, url, n, "Illegal configuration-file syntax");
            }
            int codePoint = trim.codePointAt(0);
            if (!Character.isJavaIdentifierStart(codePoint)) {
                fail(clazz, url, n, "Illegal provider-class name: " + trim);
            }
            int codePoint2;
            for (int i = Character.charCount(codePoint); i < length; i += Character.charCount(codePoint2)) {
                codePoint2 = trim.codePointAt(i);
                if (!Character.isJavaIdentifierPart(codePoint2) && codePoint2 != 46) {
                    fail(clazz, url, n, "Illegal provider-class name: " + trim);
                }
            }
            if (!set.contains(trim)) {
                list.add(trim);
                set.add(trim);
            }
        }
        return n + 1;
    }
    
    private static Iterator<String> parse(Class<?> clazz, URL url, Set<String> set) throws ServiceConfigurationError {
        InputStream openStream = null;
        BufferedReader bufferedReader = null;
        ArrayList<String> list = new ArrayList();
        try {
            openStream = url.openStream();
            bufferedReader = new BufferedReader(new InputStreamReader(openStream, "utf-8"));
            int line = 1;
            while ((line = parseLine(clazz, url, bufferedReader, line, list, set)) >= 0) {}
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException obj) {
                fail(clazz, ": " + obj);
            }
        }
        catch (IOException obj2) {
            fail(clazz, ": " + obj2);
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException obj3) {
                fail(clazz, ": " + obj3);
            }
        }
        finally {
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException obj4) {
                fail(clazz, ": " + obj4);
            }
        }
        return list.iterator();
    }
    
    public static <S> Iterator<S> providers(Class<S> clazz, ClassLoader classLoader) throws ServiceConfigurationError {
        return new Service.LazyIterator(clazz, classLoader);
    }
    
    public static <S> Iterator<S> providers(Class<S> clazz) throws ServiceConfigurationError {
        return providers(clazz, Thread.currentThread().getContextClassLoader());
    }
    
    public static <S> Iterator<S> installedProviders(Class<S> clazz) throws ServiceConfigurationError {
        ClassLoader classLoader = ClassLoader.getSystemClassLoader();
        ClassLoader classLoader2 = null;
        while (classLoader != null) {
            classLoader2 = classLoader;
            classLoader = classLoader.getParent();
        }
        return providers(clazz, classLoader2);
    }
    
    static void access$000(Class clazz, String s) throws ServiceConfigurationError {
        fail(clazz, s);
    }
    
    static Iterator access$100(Class clazz, URL url, Set set) throws ServiceConfigurationError {
        return parse(clazz, url, set);
    }
    
    static void access$200(Class clazz, String s, Throwable t) throws ServiceConfigurationError {
        fail(clazz, s, t);
    }
    
    private static class LazyIterator<S> implements Iterator<S>
    {
        Class<S> service;
        ClassLoader loader;
        Enumeration<URL> configs;
        Iterator<String> pending;
        Set<String> returned;
        String nextName;
        
        private LazyIterator(Class<S> service, ClassLoader loader) {
            this.configs = null;
            this.pending = null;
            this.returned = new TreeSet();
            this.nextName = null;
            this.service = service;
            this.loader = loader;
        }
        
        @Override
        public boolean hasNext() throws ServiceConfigurationError {
            if (this.nextName != null) {
                return true;
            }
            if (this.configs == null) {
                try {
                    String string = "META-INF/services/" + this.service.getName();
                    if (this.loader == null) {
                        this.configs = ClassLoader.getSystemResources(string);
                    }
                    else {
                        this.configs = this.loader.getResources(string);
                    }
                }
                catch (IOException obj) {
                    Service.access$000(this.service, ": " + obj);
                }
            }
            while (this.pending == null || !this.pending.hasNext()) {
                if (!this.configs.hasMoreElements()) {
                    return false;
                }
                this.pending = Service.access$100(this.service, (URL)this.configs.nextElement(), this.returned);
            }
            this.nextName = (String)this.pending.next();
            return true;
        }
        
        @Override
        public S next() throws ServiceConfigurationError {
            if (!this.hasNext()) {
                throw new NoSuchElementException();
            }
            String nextName = this.nextName;
            this.nextName = null;
            Class<Object> forName = null;
            try {
                forName = Class.forName(nextName, false, this.loader);
            }
            catch (ClassNotFoundException ex) {
                Service.access$000(this.service, "Provider " + nextName + " not found");
            }
            if (!this.service.isAssignableFrom(forName)) {
                Service.access$000(this.service, "Provider " + nextName + " not a subtype");
            }
            try {
                return this.service.cast(forName.newInstance());
            }
            catch (Throwable t) {
                Service.access$200(this.service, "Provider " + nextName + " could not be instantiated", t);
                return null;
            }
        }
        
        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
